@@ -777,7 +777,9 @@ load_end_certificate(const char *filename, struct end *dst)
 	break;
 	
     case CERT_X509_SIGNATURE:
-	select_x509cert_id(cert.u.x509, &dst->id);
+	if(dst->id.kind == ID_FROMCERT || dst->id.kind == ID_NONE) {
+	    select_x509cert_id(cert.u.x509, &dst->id);
+	}
 	
 	if (!cached_cert)
 	{
@@ -794,15 +796,15 @@ load_end_certificate(const char *filename, struct end *dst)
 	{
 	    DBG(DBG_CONTROL,
 		DBG_log("certificate is valid")
-		)
-		if (cached_cert)
-		    dst->cert = cert;
-		else
-		{
-		    add_x509_public_key(cert.u.x509, valid_until, DAL_LOCAL);
-		    dst->cert.type = cert.type;
-		    dst->cert.u.x509 = add_x509cert(cert.u.x509);
-		}
+		);
+	    if (cached_cert)
+		dst->cert = cert;
+	    else
+	    {
+		add_x509_public_key(&dst->id, cert.u.x509, valid_until, DAL_LOCAL);
+		dst->cert.type = cert.type;
+		dst->cert.u.x509 = add_x509cert(cert.u.x509);
+	    }
 	    /* if no CA is defined, use issuer as default */
 	    if (dst->ca.ptr == NULL)
 		dst->ca = dst->cert.u.x509->issuer;
@@ -1099,7 +1101,7 @@ add_connection(const struct whack_message *wm)
 	if (alg_info_ike!= NULL && alg_info_ike->alg_info_cnt==0) {
 	    loglog(RC_NOALGO
 		   , "got 0 transforms for ike=\"%s\""
-		   , wm->esp);
+		   , wm->ike);
 	    return;
 	} 
 
@@ -1141,20 +1143,15 @@ add_connection(const struct whack_message *wm)
 	{
 		DBG(DBG_CONTROL, DBG_log("from whack: got --esp=%s", wm->esp ? wm->esp: "NULL"));
 
-		if( (c->policy & POLICY_AUTHENTICATE)
-		    && (c->policy & POLICY_ENCRYPT)) {
+		if(c->policy & POLICY_ENCRYPT) {
+		    c->alg_info_esp = alg_info_esp_create_from_str(wm->esp? wm->esp : "", &ugh, FALSE);
+		} else if(c->policy & POLICY_AUTHENTICATE) {
+		    c->alg_info_esp = alg_info_ah_create_from_str(wm->esp? wm->esp : "", &ugh, FALSE);
+		} else {
 		    loglog(RC_NOALGO, "Can only do AH, or ESP, not AH+ESP\n");
 		    return;
 		}
 
-		if(c->policy & POLICY_ENCRYPT) {
-		    c->alg_info_esp = alg_info_esp_create_from_str(wm->esp? wm->esp : "", &ugh, FALSE);
-		} 
-
-		if(c->policy & POLICY_AUTHENTICATE) {
-		    c->alg_info_esp = alg_info_ah_create_from_str(wm->esp? wm->esp : "", &ugh, FALSE);
-		} 
-
 		DBG(DBG_CRYPT|DBG_CONTROL, 
 			static char buf[256]="<NULL>";
 			if (c->alg_info_esp)
@@ -1234,10 +1231,12 @@ add_connection(const struct whack_message *wm)
 	    c->spd.this.ca = c->spd.that.ca;
 
 #ifdef MODECFG
+#ifdef MODECFG_DNSWINS
 	c->modecfg_dns1 = wm->modecfg_dns1;
 	c->modecfg_dns2 = wm->modecfg_dns2;
 	c->modecfg_wins1 = wm->modecfg_wins1;
 	c->modecfg_wins2 = wm->modecfg_wins2;
+#endif
 #endif
 
 	default_end(&c->spd.this, &c->spd.that.host_addr);
@@ -2064,8 +2063,11 @@ find_host_connection2(const char *func
 		     , const ip_address *him, u_int16_t his_port, lset_t policy)
 {
     struct connection *c;
+    char mebuf[ADDRTOT_BUF], himbuf[ADDRTOT_BUF];
     DBG(DBG_CONTROLMORE,
-	DBG_log("find_host_connection called from %s policy=%s", func
+	DBG_log("find_host_connection called from %s, me=%s:%d him=%s:%d policy=%s", func
+		, (addrtot(me,  0, mebuf,  sizeof(mebuf)),mebuf),   my_port
+		, (addrtot(him, 0, himbuf, sizeof(himbuf)),himbuf), his_port
 		, bitnamesof(sa_policy_bit_names, policy)));
     c = find_host_pair_connections(__FUNCTION__, me, my_port, him, his_port);
 
@@ -2276,6 +2278,7 @@ refine_host_connection(const struct state *st, const struct id *peer_id
      *   + our ID must not change (we sent it in previous message)
      *   + our RSA key must not change (we used in in previous message)
      */
+    passert(c != NULL);
     d = c->host_pair->connections;
     for (wcpip = FALSE; ; wcpip = TRUE)
     {
@@ -2435,9 +2438,9 @@ is_virtual_net_used(struct connection *c, const ip_subnet *peer_net, const struc
 				 , enum_name(&connection_kind_names, d->kind)
 				 , buf);
 
-		    if(!kernel_ops->overlap_supported) {
+		    if(!kernel_overlap_supported()) {
 			openswan_log("Kernel method '%s' does not support overlapping IP ranges"
-				     , kernel_ops->kern_name);
+				     , kernel_if_name());
 			return TRUE;
 
 		    } else if(LIN(POLICY_OVERLAPIP, c->policy)
@@ -2689,8 +2692,8 @@ fc_try_oppo(const struct connection *c
 	    continue;
 
 	if (!(same_id(&c->spd.this.id, &d->spd.this.id)
-	&& match_id(&c->spd.that.id, &d->spd.that.id, &wildcards)
-	&& trusted_ca(c->spd.that.ca, d->spd.that.ca, &pathlen)))
+	      && match_id(&c->spd.that.id, &d->spd.that.id, &wildcards)
+	      && trusted_ca(c->spd.that.ca, d->spd.that.ca, &pathlen)))
 	    continue;
 
 	/* compare protocol and ports */
@@ -2935,11 +2938,225 @@ connection_compare_qsort(const void *a, const void *b)
 			    , *(const struct connection *const *)b);
 }
 
+static void
+show_one_sr(struct connection *c
+	    , struct spd_route *sr
+	    , int num UNUSED
+	    , char *instance)
+{
+    char topo[CONN_BUF_LEN];
+    char srcip[ADDRTOT_BUF], dstip[ADDRTOT_BUF];
+    char thissemi[3+sizeof("myup=")];
+    char thatsemi[3+sizeof("hisup=")];
+#ifdef XAUTH
+    char thisxauthsemi[XAUTH_USERNAME_LEN+sizeof("myxauthuser=")];
+    char thatxauthsemi[XAUTH_USERNAME_LEN+sizeof("hisxauthuser=")];
+#endif
+    char thiscertsemi[3+sizeof("mycert=")+PATH_MAX];
+    char thatcertsemi[3+sizeof("hiscert=")+PATH_MAX];
+    char *thisup, *thatup;
+    
+    (void) format_connection(topo, sizeof(topo), c, sr);
+    whack_log(RC_COMMENT, "\"%s\"%s: %s; %s; eroute owner: #%lu"
+	      , c->name, instance, topo
+	      , enum_name(&routing_story, sr->routing)
+	      , sr->eroute_owner);
+    if(addrbytesptr(&c->spd.this.host_srcip, NULL) == 0
+       || isanyaddr(&c->spd.this.host_srcip)) {
+	strcpy(srcip, "unset");
+    } else {
+	addrtot(&sr->this.host_srcip, 0, srcip, sizeof(srcip));
+    }
+    if(addrbytesptr(&c->spd.that.host_srcip, NULL) == 0
+       || isanyaddr(&c->spd.that.host_srcip)) {
+	strcpy(dstip, "unset");
+    } else {
+	addrtot(&sr->that.host_srcip, 0, dstip, sizeof(dstip));
+    }
+    
+    thissemi[0]='\0';
+    thisup=thissemi;
+    if(sr->this.updown) {
+	thissemi[0]=';';
+	thissemi[1]=' ';
+	thissemi[2]='\0';
+	strcat(thissemi, "myup=");
+	thisup=sr->this.updown;
+    }
+    
+    thatsemi[0]='\0';
+    thatup=thatsemi;
+    if(sr->that.updown) {
+	thatsemi[0]=';';
+	thatsemi[1]=' ';
+	thatsemi[2]='\0';
+	strcat(thatsemi, "hisup=");
+	thatup=sr->that.updown;
+    }
+    
+    thiscertsemi[0]='\0';
+    if(sr->this.cert_filename) {
+	snprintf(thiscertsemi, sizeof(thiscertsemi)-1
+		 , "; mycert=%s"
+		 , sr->this.cert_filename);
+    }
+    
+    thatcertsemi[0]='\0';
+    if(sr->that.cert_filename) {
+	snprintf(thatcertsemi, sizeof(thatcertsemi)-1
+		 , "; hiscert=%s"
+		 , sr->that.cert_filename);
+    }
+
+    whack_log(RC_COMMENT, "\"%s\"%s:     myip=%s; hisip=%s%s%s%s%s%s%s;"
+	      , c->name, instance, srcip, dstip
+	      , thissemi, thisup
+	      , thatsemi, thatup
+	      , thiscertsemi
+	      , thatcertsemi);
+    
+#ifdef XAUTH
+    if(sr->this.xauth_name || sr->that.xauth_name) {
+	thisxauthsemi[0]='\0';
+	if(sr->this.xauth_name) {
+	    snprintf(thisxauthsemi, sizeof(thisxauthsemi)-1
+		     , "myxauthuser=%s; "
+		     , sr->this.xauth_name);
+	}
+	
+	thatxauthsemi[0]='\0';
+	if(sr->that.xauth_name) {
+	    snprintf(thatxauthsemi, sizeof(thatxauthsemi)-1
+		     , "hisxauthuser=%s; "
+		     , sr->that.xauth_name);
+	}
+	whack_log(RC_COMMENT, "\"%s\"%s:     xauth info: %s%s"
+		  , c->name, instance
+		  , thisxauthsemi
+		  , thatxauthsemi);
+    }
+#endif
+}
+
+void 
+show_one_connection(struct connection *c)
+{
+    const char *ifn;
+    char instance[1 + 10 + 1];
+    char prio[POLICY_PRIO_BUF];
+    
+    ifn = oriented(*c)? c->interface->ip_dev->id_rname : "";
+
+    instance[0] = '\0';
+    if (c->kind == CK_INSTANCE && c->instance_serial != 0)
+	snprintf(instance, sizeof(instance), "[%lu]", c->instance_serial);
+    
+    /* show topology */
+    {
+	struct spd_route *sr = &c->spd;
+	int num=0;
+	
+	while (sr != NULL)
+	{
+	    show_one_sr(c, sr, num, instance);
+	    sr = sr->next;
+	    num++;
+	}
+    }
+    
+    /* show CAs */
+    if (c->spd.this.ca.ptr != NULL || c->spd.that.ca.ptr != NULL)
+    {
+	char this_ca[IDTOA_BUF], that_ca[IDTOA_BUF];
+	
+	dntoa_or_null(this_ca, IDTOA_BUF, c->spd.this.ca, "%any");
+	dntoa_or_null(that_ca, IDTOA_BUF, c->spd.that.ca, "%any");
+	
+	whack_log(RC_COMMENT
+		  , "\"%s\"%s:   CAs: '%s'...'%s'"
+		  , c->name
+		  , instance
+		  , this_ca
+		  , that_ca);
+    }
+    
+    whack_log(RC_COMMENT
+	      , "\"%s\"%s:   ike_life: %lus; ipsec_life: %lus;"
+	      " rekey_margin: %lus; rekey_fuzz: %lu%%; keyingtries: %lu"
+	      , c->name
+	      , instance
+	      , (unsigned long) c->sa_ike_life_seconds
+	      , (unsigned long) c->sa_ipsec_life_seconds
+	      , (unsigned long) c->sa_rekey_margin
+	      , (unsigned long) c->sa_rekey_fuzz
+	      , (unsigned long) c->sa_keying_tries);
+
+    if (c->policy_next)
+    {
+	whack_log(RC_COMMENT
+		  , "\"%s\"%s:   policy_next: %s"
+		  , c->name, instance, c->policy_next->name);
+    }
+    
+    /* Note: we display key_from_DNS_on_demand as if policy [lr]KOD */
+    fmt_policy_prio(c->prio, prio);
+    whack_log(RC_COMMENT
+	      , "\"%s\"%s:   policy: %s%s%s; prio: %s; interface: %s; "
+	      , c->name
+	      , instance
+	      , prettypolicy(c->policy)
+	      , c->spd.this.key_from_DNS_on_demand? "+lKOD" : ""
+	      , c->spd.that.key_from_DNS_on_demand? "+rKOD" : ""
+	      , prio
+	      , ifn);
+    
+    /* slightly complicated stuff to avoid extra crap */
+    if(c->dpd_timeout > 0 || DBGP(DBG_DPD)) {
+	whack_log(RC_COMMENT
+		  , "\"%s\"%s:   dpd: %s; delay:%lu; timeout:%lu; "
+		  , c->name
+		  , instance
+		  , enum_name(&dpd_action_names, c->dpd_action)
+		  , (unsigned long)c->dpd_delay
+		  , (unsigned long)c->dpd_timeout);
+    }
+
+    if(c->extra_debugging) {
+	whack_log(RC_COMMENT, "\"%s\"%s:   debug: %s"
+		  , c->name
+		  , instance
+		  , bitnamesof(debug_bit_names
+			       , c->extra_debugging));
+    }
+    
+    whack_log(RC_COMMENT
+	      , "\"%s\"%s:   newest ISAKMP SA: #%ld; newest IPsec SA: #%ld; "
+	      , c->name
+	      , instance
+	      , c->newest_isakmp_sa
+	      , c->newest_ipsec_sa);
+    
+    if(c->connalias) {
+	whack_log(RC_COMMENT
+		  , "\"%s\"%s:   aliases: %s\n"
+		  , c->name
+		  , instance
+		  , c->connalias);
+    }
+    
+#ifdef IKE_ALG
+    ike_alg_show_connection(c, instance);
+#endif
+#ifdef KERNEL_ALG
+    kernel_alg_show_connection(c, instance);
+#endif
+}
+
 void
 show_connections_status(void)
 {
-    struct connection *c;
     int count, i;
+    struct connection *c;
     struct connection **array;
 
 
@@ -2962,208 +3179,7 @@ show_connections_status(void)
 
     for (i=0; i<count; i++)
     {
-	const char *ifn;
-	char instance[1 + 10 + 1];
-	char prio[POLICY_PRIO_BUF];
-
-	c = array[i];
-
-	ifn = oriented(*c)? c->interface->ip_dev->id_rname : "";
-
-	instance[0] = '\0';
-	if (c->kind == CK_INSTANCE && c->instance_serial != 0)
-	    snprintf(instance, sizeof(instance), "[%lu]", c->instance_serial);
-
-	/* show topology */
-	{
-	    char topo[CONN_BUF_LEN];
-	    struct spd_route *sr = &c->spd;
-	    int num=0;
-
-	    while (sr != NULL)
-	    {
-		char srcip[ADDRTOT_BUF], dstip[ADDRTOT_BUF];
-		char thissemi[3+sizeof("myup=")];
-		char thatsemi[3+sizeof("hisup=")];
-#ifdef XAUTH
-		char thisxauthsemi[XAUTH_USERNAME_LEN+sizeof("myxauthuser=")];
-		char thatxauthsemi[XAUTH_USERNAME_LEN+sizeof("hisxauthuser=")];
-#endif
-		char thiscertsemi[3+sizeof("mycert=")+PATH_MAX];
-		char thatcertsemi[3+sizeof("hiscert=")+PATH_MAX];
-		char *thisup, *thatup;
-
-		(void) format_connection(topo, sizeof(topo), c, sr);
-		whack_log(RC_COMMENT, "\"%s\"%s: %s; %s; eroute owner: #%lu"
-			  , c->name, instance, topo
-			  , enum_name(&routing_story, sr->routing)
-			  , sr->eroute_owner);
-		if(addrbytesptr(&c->spd.this.host_srcip, NULL) == 0
-		   || isanyaddr(&c->spd.this.host_srcip)) {
-		    strcpy(srcip, "unset");
-		} else {
-		    addrtot(&sr->this.host_srcip, 0, srcip, sizeof(srcip));
-		}
-		if(addrbytesptr(&c->spd.that.host_srcip, NULL) == 0
-		   || isanyaddr(&c->spd.that.host_srcip)) {
-		    strcpy(dstip, "unset");
-		} else {
-		    addrtot(&sr->that.host_srcip, 0, dstip, sizeof(dstip));
-		}
-
-		thissemi[0]='\0';
-		thisup=thissemi;
-		if(sr->this.updown) {
-		    thissemi[0]=';';
-		    thissemi[1]=' ';
-		    thissemi[2]='\0';
-		    strcat(thissemi, "myup=");
-		    thisup=sr->this.updown;
-		}
-		
-		thatsemi[0]='\0';
-		thatup=thatsemi;
-		if(sr->that.updown) {
-		    thatsemi[0]=';';
-		    thatsemi[1]=' ';
-		    thatsemi[2]='\0';
-		    strcat(thatsemi, "hisup=");
-		    thatup=sr->that.updown;
-		}
-		
-		thiscertsemi[0]='\0';
-		if(sr->this.cert_filename) {
-		    snprintf(thiscertsemi, sizeof(thiscertsemi)-1
-			     , "; mycert=%s"
-			     , sr->this.cert_filename);
-		}
-
-		thatcertsemi[0]='\0';
-		if(sr->that.cert_filename) {
-		    snprintf(thatcertsemi, sizeof(thatcertsemi)-1
-			     , "; hiscert=%s"
-			     , sr->that.cert_filename);
-		}
-
-		whack_log(RC_COMMENT, "\"%s\"%s:     myip=%s; hisip=%s%s%s%s%s%s%s;"
-			  , c->name, instance, srcip, dstip
-			  , thissemi, thisup
-			  , thatsemi, thatup
-			  , thiscertsemi
-			  , thatcertsemi);
-
-#ifdef XAUTH
-		if(sr->this.xauth_name || sr->that.xauth_name) {
-		    thisxauthsemi[0]='\0';
-		    if(sr->this.xauth_name) {
-			snprintf(thisxauthsemi, sizeof(thisxauthsemi)-1
-				 , "myxauthuser=%s; "
-				 , sr->this.xauth_name);
-		    }
-		    
-		    thatxauthsemi[0]='\0';
-		    if(sr->that.xauth_name) {
-			snprintf(thatxauthsemi, sizeof(thatxauthsemi)-1
-				 , "hisxauthuser=%s; "
-				 , sr->that.xauth_name);
-		    }
-		    whack_log(RC_COMMENT, "\"%s\"%s:     xauth info: %s%s"
-			      , c->name, instance
-			      , thisxauthsemi
-			      , thatxauthsemi);
-		}
-#endif
-		sr = sr->next;
-		num++;
-	    }
-	}
-
-	/* show CAs */
-	if (c->spd.this.ca.ptr != NULL || c->spd.that.ca.ptr != NULL)
-	{
-	    char this_ca[IDTOA_BUF], that_ca[IDTOA_BUF];
-
-	    dntoa_or_null(this_ca, IDTOA_BUF, c->spd.this.ca, "%any");
-	    dntoa_or_null(that_ca, IDTOA_BUF, c->spd.that.ca, "%any");
-
-	    whack_log(RC_COMMENT
-		, "\"%s\"%s:   CAs: '%s'...'%s'"
-		, c->name
-		, instance
-		, this_ca
-		, that_ca);
-	}
-
-	whack_log(RC_COMMENT
-	    , "\"%s\"%s:   ike_life: %lus; ipsec_life: %lus;"
-	    " rekey_margin: %lus; rekey_fuzz: %lu%%; keyingtries: %lu"
-	    , c->name
-	    , instance
-	    , (unsigned long) c->sa_ike_life_seconds
-	    , (unsigned long) c->sa_ipsec_life_seconds
-	    , (unsigned long) c->sa_rekey_margin
-	    , (unsigned long) c->sa_rekey_fuzz
-	    , (unsigned long) c->sa_keying_tries);
-
-	if (c->policy_next)
-	{
-	    whack_log(RC_COMMENT
-		      , "\"%s\"%s:   policy_next: %s"
-		      , c->name, instance, c->policy_next->name);
-	}
-
-	/* Note: we display key_from_DNS_on_demand as if policy [lr]KOD */
-	fmt_policy_prio(c->prio, prio);
-	whack_log(RC_COMMENT
-	    , "\"%s\"%s:   policy: %s%s%s; prio: %s; interface: %s; "
-	    , c->name
-	    , instance
-	    , prettypolicy(c->policy)
-	    , c->spd.this.key_from_DNS_on_demand? "+lKOD" : ""
-	    , c->spd.that.key_from_DNS_on_demand? "+rKOD" : ""
-	    , prio
-	    , ifn);
-
-	/* slightly complicated stuff to avoid extra crap */
-	if(c->dpd_timeout > 0 || DBGP(DBG_DPD)) {
-	    whack_log(RC_COMMENT
-		      , "\"%s\"%s:   dpd: %s; delay:%lu; timeout:%lu; "
-		      , c->name
-		      , instance
-		      , enum_name(&dpd_action_names, c->dpd_action)
-		      , (unsigned long)c->dpd_delay
-		      , (unsigned long)c->dpd_timeout);
-	}
-
-	if(c->extra_debugging) {
-	    whack_log(RC_COMMENT, "\"%s\"%s:   debug: %s"
-		      , c->name
-		      , instance
-		      , bitnamesof(debug_bit_names
-				   , c->extra_debugging));
-	}
-
-	whack_log(RC_COMMENT
-	    , "\"%s\"%s:   newest ISAKMP SA: #%ld; newest IPsec SA: #%ld; "
-	    , c->name
-	    , instance
-	    , c->newest_isakmp_sa
-	    , c->newest_ipsec_sa);
-
-	if(c->connalias) {
-	    whack_log(RC_COMMENT
-		      , "\"%s\"%s:   aliases: %s\n"
-		      , c->name
-		      , instance
-		      , c->connalias);
-	}
-
-#ifdef IKE_ALG
-	ike_alg_show_connection(c, instance);
-#endif
-#ifdef KERNEL_ALG
-	kernel_alg_show_connection(c, instance);
-#endif
+	show_one_connection(array[i]);
     }
     pfree(array);
 }
