@@ -105,7 +105,12 @@ out:
 }
 
 static void
-__alg_info_ike_add (struct alg_info_ike *alg_info, int ealg_id, unsigned ek_bits, int aalg_id, unsigned ak_bits, int modp_id)
+__alg_info_ike_add (struct alg_info_ike *alg_info
+		    , int ealg_id
+		    , unsigned ek_bits
+		    , int aalg_id
+		    , unsigned ak_bits
+		    , unsigned modp_id)
 {
 	struct ike_info *ike_info=alg_info->ike;
 	unsigned cnt=alg_info->alg_info_cnt, i;
@@ -492,14 +497,15 @@ kernel_alg_db_add(struct db_context *db_ctx
 struct db_context * 
 kernel_alg_db_new(struct alg_info_esp *alg_info, lset_t policy, bool logit)
 {
-	int ealg_i, aalg_i, tn=0;
+    int ealg_i, aalg_i;
+    unsigned int tn=0;
 	int i;
 	const struct esp_info *esp_info;
 	struct esp_info tmp_esp_info;
 	struct db_context *ctx_new=NULL;
 	struct db_trans *t;
 	struct db_prop  *prop;
-	int trans_cnt;
+	unsigned int trans_cnt;
 	bool success = TRUE;
 	int protoid;
 
@@ -573,8 +579,8 @@ kernel_alg_db_new(struct alg_info_esp *alg_info, lset_t policy, bool logit)
 			"attrs[0].type=%d, attrs[0].val=%d"
 			, tn
 			, t[tn].transid, t[tn].attr_cnt
-			, t[tn].attrs ? t[tn].attrs[0].type : -1
-			, t[tn].attrs ? t[tn].attrs[0].val : -1
+			, t[tn].attrs ? t[tn].attrs[0].type.ipsec : 255
+			, t[tn].attrs ? t[tn].attrs[0].val : 255
 			));
 	}
 	prop->trans_cnt = tn;
@@ -663,7 +669,7 @@ void kernel_alg_show_status(void)
 void
 kernel_alg_show_connection(struct connection *c, const char *instance)
 {
-	char buf[1024];
+	char buf[256];
 	struct state *st;
 	const char *satype;
 
@@ -697,10 +703,11 @@ kernel_alg_show_connection(struct connection *c, const char *instance)
 		, "\"%s\"%s:   %s algorithm newest: %s_%d-%s; pfsgroup=%s"
 		, c->name
 			  , instance, satype
-		, enum_show(&esp_transformid_names, st->st_esp.attrs.transid)
+		, enum_show(&esp_transformid_names
+			    ,st->st_esp.attrs.transattrs.encrypt)
 		+4 /* strlen("ESP_") */
-		, st->st_esp.attrs.key_len
-		, enum_show(&auth_alg_names, st->st_esp.attrs.auth)+
+		, st->st_esp.attrs.transattrs.enckeylen
+		, enum_show(&auth_alg_names, st->st_esp.attrs.transattrs.integ_hash)+
 		+15 /* strlen("AUTH_ALGORITHM_") */
 		, c->policy & POLICY_PFS ?
 			c->alg_info_esp->esp_pfsgroup ?
@@ -716,7 +723,7 @@ kernel_alg_show_connection(struct connection *c, const char *instance)
 		, "\"%s\"%s:   %s algorithm newest: %s; pfsgroup=%s"
 		, c->name
 			  , instance, satype
-		, enum_show(&auth_alg_names, st->st_esp.attrs.auth)+
+		, enum_show(&auth_alg_names, st->st_esp.attrs.transattrs.integ_hash)+
 		+15 /* strlen("AUTH_ALGORITHM_") */
 		, c->policy & POLICY_PFS ?
 			c->alg_info_esp->esp_pfsgroup ?
@@ -737,9 +744,25 @@ kernel_alg_makedb(lset_t policy, struct alg_info_esp *ei, bool logit)
     struct db_prop_conj pc;
     struct db_sa t, *n;
 
+    memset(&t, 0, sizeof(t));
+
     if(ei == NULL) {
-	DBG(DBG_CONTROL, DBG_log("empty esp_info, returning empty"));
-	return NULL;
+	struct db_sa *sadb;
+	lset_t pm = POLICY_ENCRYPT | POLICY_AUTHENTICATE;
+
+#if 0
+y	if (can_do_IPcomp)
+	    pm |= POLICY_COMPRESS;
+#endif
+
+	sadb = &ipsec_sadb[(policy & pm) >> POLICY_IPSEC_SHIFT];
+
+	/* make copy, to keep from freeing the static policies */
+	sadb = sa_copy_sa(sadb, 0);
+	sadb->parentSA = FALSE;
+
+	DBG(DBG_CONTROL, DBG_log("empty esp_info, returning defaults"));
+	return sadb;
     }
     
     dbnew=kernel_alg_db_new(ei, policy, logit);
@@ -764,6 +787,7 @@ kernel_alg_makedb(lset_t policy, struct alg_info_esp *ei, bool logit)
 
     /* make a fresh copy */
     n = sa_copy_sa(&t, 0);
+    n->parentSA = FALSE;
     
     db_destroy(dbnew);
 
